#include Bluepad32.h
#include Adafruit_VL53L0X.h

 Pin definitions
const int motor1Pin1 = 9;  IN1 for Motor 1
const int motor1Pin2 = 8;  IN2 for Motor 1
const int motor2Pin1 = 7;  IN3 for Motor 2
const int motor2Pin2 = 44;  IN4 for Motor 2
const int enablePin1 = 2;  ENA for Motor 1
const int enablePin2 = 3;  ENB for Motor 2

const int pwmChannel1 = 0;
const int pwmChannel2 = 1;
const int pwmFrequency = 5000;
const int pwmResolution = 8;

const int stopDistance = 100;  Distance threshold to stop the motors (in mm)

ControllerPtr myControllers[BP32_MAX_GAMEPADS];
Adafruit_VL53L0X lox = Adafruit_VL53L0X();  Initialize the sensor

void onConnectedController(ControllerPtr ctl) {
    bool foundEmptySlot = false;
    for (int i = 0; i  BP32_MAX_GAMEPADS; i++) {
        if (myControllers[i] == nullptr) {
            Serial.printf(CALLBACK Controller is connected, index=%dn, i);
            ControllerProperties properties = ctl-getProperties();
            Serial.printf(Controller model %s, VID=0x%04x, PID=0x%04xn, ctl-getModelName().c_str(), properties.vendor_id, properties.product_id);
            myControllers[i] = ctl;
            foundEmptySlot = true;
            break;
        }
    }
    if (!foundEmptySlot) {
        Serial.println(CALLBACK Controller connected, but could not found empty slot);
    }
}

void onDisconnectedController(ControllerPtr ctl) {
    bool foundController = false;

    for (int i = 0; i  BP32_MAX_GAMEPADS; i++) {
        if (myControllers[i] == ctl) {
            Serial.printf(CALLBACK Controller disconnected from index=%dn, i);
            myControllers[i] = nullptr;
            foundController = true;
            break;
        }
    }

    if (!foundController) {
        Serial.println(CALLBACK Controller disconnected, but not found in myControllers);
    }
}

void dumpGamepad(ControllerPtr ctl) {
    Serial.printf(
        idx=%d, dpad 0x%02x, buttons 0x%04x, axis L %4d, %4d, axis R %4d, %4d, brake %4d, throttle %4d, 
        misc 0x%02x, gyro x%6d y%6d z%6d, accel x%6d y%6d z%6dn,
        ctl-index(),         Controller Index
        ctl-dpad(),          D-pad
        ctl-buttons(),       bitmask of pressed buttons
        ctl-axisX(),         (-511 - 512) left X Axis
        ctl-axisY(),         (-511 - 512) left Y axis
        ctl-axisRX(),        (-511 - 512) right X axis
        ctl-axisRY(),        (-511 - 512) right Y axis
        ctl-brake(),         (0 - 1023) brake button
        ctl-throttle(),      (0 - 1023) throttle (AKA gas) button
        ctl-miscButtons(),   bitmask of pressed misc buttons
        ctl-gyroX(),         Gyro X
        ctl-gyroY(),         Gyro Y
        ctl-gyroZ(),         Gyro Z
        ctl-accelX(),        Accelerometer X
        ctl-accelY(),        Accelerometer Y
        ctl-accelZ()         Accelerometer Z
    );
}

void stopMotors() {
    digitalWrite(motor1Pin1, LOW);
    digitalWrite(motor1Pin2, LOW);
    ledcWrite(pwmChannel1, 0);
    digitalWrite(motor2Pin1, LOW);
    digitalWrite(motor2Pin2, LOW);
    ledcWrite(pwmChannel2, 0);
    Serial.println(Motors stopped.);
}

bool checkDistance() {
    uint16_t distance = lox.readRange();
    Serial.print(Distance in mm );
    Serial.println(distance);

    if (distance  stopDistance) {
        stopMotors();
        Serial.println(Object detected! Motors stopped.);
        return false;
    }
    return true;
}

void processGamepad(ControllerPtr ctl) {
     Ensure we only process gamepad input if distance is safe
    if (!checkDistance()) {
        return;
    }

     Print out raw values for debugging
    Serial.printf(Raw throttle %d, Raw brake %dn, ctl-throttle(), ctl-brake());
    Serial.printf(Raw axisRX %d, Raw axisRY %dn, ctl-axisRX(), ctl-axisRY());

     Normalize throttle and brake values (0-255)
    int throttle = ctl-throttle()  4;
    int brake = ctl-brake()  4;

    Serial.printf(Normalized Throttle %d, Normalized Brake %dn, throttle, brake);

     Calculate speed based on throttle and brake
    int speed = throttle - brake;

     Normalize direction value (-255 to 255)
    int direction = map(ctl-axisRX(), -512, 512, -255, 255);

    Serial.printf(Calculated Speed %d, Calculated Direction %dn, speed, direction);

     Calculate motor speeds
    int motorSpeed1 = speed - direction;
    int motorSpeed2 = speed + direction;

     Ensure motor speeds are within range
    motorSpeed1 = constrain(motorSpeed1, -255, 255);
    motorSpeed2 = constrain(motorSpeed2, -255, 255);

    Serial.printf(MotorSpeed1 %d, MotorSpeed2 %dn, motorSpeed1, motorSpeed2);

     Set motor directions and speeds
    if (motorSpeed1  0) {
        digitalWrite(motor1Pin1, HIGH);
        digitalWrite(motor1Pin2, LOW);
        ledcWrite(pwmChannel1, motorSpeed1);
    } else {
        digitalWrite(motor1Pin1, LOW);
        digitalWrite(motor1Pin2, HIGH);
        ledcWrite(pwmChannel1, -motorSpeed1);
    }

    if (motorSpeed2  0) {
        digitalWrite(motor2Pin1, HIGH);
        digitalWrite(motor2Pin2, LOW);
        ledcWrite(pwmChannel2, motorSpeed2);
    } else {
        digitalWrite(motor2Pin1, LOW);
        digitalWrite(motor2Pin2, HIGH);
        ledcWrite(pwmChannel2, -motorSpeed2);
    }

    dumpGamepad(ctl);
}

void processControllers() {
    for (auto myController  myControllers) {
        if (myController && myController-isConnected() && myController-hasData()) {
            if (myController-isGamepad()) {
                processGamepad(myController);
            }
        }
    }
}

 Arduino setup function. Runs in CPU 1
void setup() {
    Serial.begin(115200);
    Serial.printf(Firmware %sn, BP32.firmwareVersion());
    const uint8_t addr = BP32.localBdAddress();
    Serial.printf(BD Addr %2X%2X%2X%2X%2X%2Xn, addr[0], addr[1], addr[2], addr[3], addr[4], addr[5]);

     Motor driver pin setup
    pinMode(motor1Pin1, OUTPUT);
    pinMode(motor1Pin2, OUTPUT);
    pinMode(motor2Pin1, OUTPUT);
    pinMode(motor2Pin2, OUTPUT);
    pinMode(enablePin1, OUTPUT);
    pinMode(enablePin2, OUTPUT);

     Setup PWM for motor control
    ledcSetup(pwmChannel1, pwmFrequency, pwmResolution);
    ledcAttachPin(enablePin1, pwmChannel1);
    ledcSetup(pwmChannel2, pwmFrequency, pwmResolution);
    ledcAttachPin(enablePin2, pwmChannel2);

     Setup the Bluepad32 callbacks
    BP32.setup(&onConnectedController, &onDisconnectedController);

     forgetBluetoothKeys() should be called when the user performs
     a device factory reset, or similar.
     Calling forgetBluetoothKeys in setup() just as an example.
     Forgetting Bluetooth keys prevents paired gamepads to reconnect.
     But it might also fix some connection  re-connection issues.
    BP32.forgetBluetoothKeys();

     Enables mouse  touchpad support for gamepads that support them.
     When enabled, controllers like DualSense and DualShock4 generate two connected devices
     - First one the gamepad
     - Second one, which is a virtual device, is a mouse.
     By default, it is disabled.
    BP32.enableVirtualDevice(false);

     Initialize the VL53L0X sensor
    if (!lox.begin()) {
        Serial.println(F(Failed to boot VL53L0X));
        while (1);
    }
    lox.startRangeContinuous();
}

 Arduino loop function. Runs in CPU 1.
void loop() {
     This call fetches all the controllers' data.
     Call this function in your main loop.
    bool dataUpdated = BP32.update();
    if (dataUpdated) {
        processControllers();
    }

     Continuously check the distance sensor and stop motors if necessary
    checkDistance();

     The main loop must have some kind of yield to lower priority task event.
     Otherwise, the watchdog will get triggered.
     If your main loop doesn't have one, just add a simple `vTaskDelay(1)`.
     Detailed info here httpsstackoverflow.coma60133039
    vTaskDelay(1);
}
